# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import NamedTuple
from http import HTTPStatus
from json import JSONDecodeError
import asyncio
import hashlib
import hmac
import random
import string

from aiohttp import web
from yarl import URL

from mautrix.api import Method, Path, SynapseAdminPath
from mautrix.client import ClientAPI
from mautrix.errors import MatrixRequestError
from mautrix.types import LoginResponse, LoginType

from .base import get_config, routes
from .client import _create_client, _create_or_update_client
from .responses import resp


def known_homeservers() -> dict[str, dict[str, str]]:
    return get_config()["homeservers"]


@routes.get("/client/auth/servers")
async def get_known_servers(_: web.Request) -> web.Response:
    return web.json_response({key: value["url"] for key, value in known_homeservers().items()})


class AuthRequestInfo(NamedTuple):
    server_name: str
    client: ClientAPI
    secret: str
    username: str
    password: str
    user_type: str
    device_name: str
    update_client: bool
    sso: bool


truthy_strings = ("1", "true", "yes")


async def read_client_auth_request(
    request: web.Request,
) -> tuple[AuthRequestInfo | None, web.Response | None]:
    server_name = request.match_info.get("server", None)
    server = known_homeservers().get(server_name, None)
    if not server:
        return None, resp.server_not_found
    try:
        body = await request.json()
    except JSONDecodeError:
        return None, resp.body_not_json
    sso = request.query.get("sso", "").lower() in truthy_strings
    try:
        username = body["username"]
        password = body["password"]
    except KeyError:
        if not sso:
            return None, resp.username_or_password_missing
        username = password = None
    try:
        base_url = server["url"]
    except KeyError:
        return None, resp.invalid_server
    return (
        AuthRequestInfo(
            server_name=server_name,
            client=ClientAPI(base_url=base_url),
            secret=server.get("secret"),
            username=username,
            password=password,
            user_type=body.get("user_type", "bot"),
            device_name=body.get("device_name", "Maubot"),
            update_client=request.query.get("update_client", "").lower() in truthy_strings,
            sso=sso,
        ),
        None,
    )


def generate_mac(
    secret: str,
    nonce: str,
    username: str,
    password: str,
    admin: bool = False,
    user_type: str = None,
) -> str:
    mac = hmac.new(key=secret.encode("utf-8"), digestmod=hashlib.sha1)
    mac.update(nonce.encode("utf-8"))
    mac.update(b"\x00")
    mac.update(username.encode("utf-8"))
    mac.update(b"\x00")
    mac.update(password.encode("utf-8"))
    mac.update(b"\x00")
    mac.update(b"admin" if admin else b"notadmin")
    if user_type is not None:
        mac.update(b"\x00")
        mac.update(user_type.encode("utf8"))
    return mac.hexdigest()


@routes.post("/client/auth/{server}/register")
async def register(request: web.Request) -> web.Response:
    req, err = await read_client_auth_request(request)
    if err is not None:
        return err
    if req.sso:
        return resp.registration_no_sso
    elif not req.secret:
        return resp.registration_secret_not_found
    path = SynapseAdminPath.v1.register
    res = await req.client.api.request(Method.GET, path)
    content = {
        "nonce": res["nonce"],
        "username": req.username,
        "password": req.password,
        "admin": False,
        "user_type": req.user_type,
    }
    content["mac"] = generate_mac(**content, secret=req.secret)
    try:
        raw_res = await req.client.api.request(Method.POST, path, content=content)
    except MatrixRequestError as e:
        return web.json_response(
            {
                "errcode": e.errcode,
                "error": e.message,
                "http_status": e.http_status,
            },
            status=HTTPStatus.INTERNAL_SERVER_ERROR,
        )
    login_res = LoginResponse.deserialize(raw_res)
    if req.update_client:
        return await _create_client(
            login_res.user_id,
            {
                "homeserver": str(req.client.api.base_url),
                "access_token": login_res.access_token,
                "device_id": login_res.device_id,
            },
        )
    return web.json_response(login_res.serialize())


@routes.post("/client/auth/{server}/login")
async def login(request: web.Request) -> web.Response:
    req, err = await read_client_auth_request(request)
    if err is not None:
        return err
    if req.sso:
        return await _do_sso(req)
    else:
        return await _do_login(req)


async def _do_sso(req: AuthRequestInfo) -> web.Response:
    flows = await req.client.get_login_flows()
    if not flows.supports_type(LoginType.SSO):
        return resp.sso_not_supported
    waiter_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=16))
    cfg = get_config()
    public_url = (
        URL(cfg["server.public_url"])
        / "_matrix/maubot/v1/client/auth_external_sso/complete"
        / waiter_id
    )
    sso_url = req.client.api.base_url.with_path(str(Path.v3.login.sso.redirect)).with_query(
        {"redirectUrl": str(public_url)}
    )
    sso_waiters[waiter_id] = req, asyncio.get_running_loop().create_future()
    return web.json_response({"sso_url": str(sso_url), "id": waiter_id})


async def _do_login(req: AuthRequestInfo, login_token: str | None = None) -> web.Response:
    device_id = "".join(random.choices(string.ascii_uppercase + string.digits, k=8))
    device_id = f"maubot_{device_id}"
    try:
        if req.sso:
            res = await req.client.login(
                token=login_token,
                login_type=LoginType.TOKEN,
                device_id=device_id,
                store_access_token=False,
                initial_device_display_name=req.device_name,
            )
        else:
            res = await req.client.login(
                identifier=req.username,
                login_type=LoginType.PASSWORD,
                password=req.password,
                device_id=device_id,
                initial_device_display_name=req.device_name,
                store_access_token=False,
            )
    except MatrixRequestError as e:
        return web.json_response(
            {
                "errcode": e.errcode,
                "error": e.message,
            },
            status=e.http_status,
        )
    if req.update_client:
        return await _create_or_update_client(
            res.user_id,
            {
                "homeserver": str(req.client.api.base_url),
                "access_token": res.access_token,
                "device_id": res.device_id,
            },
            is_login=True,
        )
    return web.json_response(res.serialize())


sso_waiters: dict[str, tuple[AuthRequestInfo, asyncio.Future]] = {}


@routes.post("/client/auth/{server}/sso/{id}/wait")
async def wait_sso(request: web.Request) -> web.Response:
    waiter_id = request.match_info["id"]
    req, fut = sso_waiters[waiter_id]
    try:
        login_token = await fut
    finally:
        sso_waiters.pop(waiter_id, None)
    return await _do_login(req, login_token)


@routes.get("/client/auth_external_sso/complete/{id}")
async def complete_sso(request: web.Request) -> web.Response:
    try:
        _, fut = sso_waiters[request.match_info["id"]]
    except KeyError:
        return web.Response(status=404, text="Invalid session ID\n")
    if fut.cancelled():
        return web.Response(status=200, text="The login was cancelled from the Maubot client\n")
    elif fut.done():
        return web.Response(status=200, text="The login token was already received\n")
    try:
        fut.set_result(request.query["loginToken"])
    except KeyError:
        return web.Response(status=400, text="Missing loginToken query parameter\n")
    except asyncio.InvalidStateError:
        return web.Response(status=500, text="Invalid state\n")
    return web.Response(
        status=200,
        text="Login token received, please return to your Maubot client. "
        "This tab can be closed.\n",
    )



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import json
import os

from colorama import Fore
from yarl import URL
import aiohttp

from ..cliq import cliq
from ..config import config, save_config


@cliq.command(help="Log in to a Maubot instance")
@cliq.option(
    "-u",
    "--username",
    help="The username of your account",
    default=os.environ.get("USER", None),
    required=True,
)
@cliq.option(
    "-p", "--password", help="The password to your account", inq_type="password", required=True
)
@cliq.option(
    "-s",
    "--server",
    help="The server to log in to",
    default="http://localhost:29316",
    required=True,
)
@cliq.option(
    "-a",
    "--alias",
    help="Alias to reference the server without typing the full URL",
    default="",
    required=False,
)
@cliq.with_http
async def login(
    server: str, username: str, password: str, alias: str, sess: aiohttp.ClientSession
) -> None:
    data = {
        "username": username,
        "password": password,
    }
    url = URL(server) / "_matrix/maubot/v1/auth/login"
    async with sess.post(url, json=data) as resp:
        if resp.status == 200:
            data = await resp.json()
            config["servers"][server] = data["token"]
            if not config["default_server"]:
                print(Fore.CYAN, "Setting", server, "as the default server")
                config["default_server"] = server
            if alias:
                config["aliases"][alias] = server
            save_config()
            print(Fore.GREEN + "Logged in successfully")
        else:
            try:
                err = (await resp.json())["error"]
            except (json.JSONDecodeError, KeyError):
                err = await resp.text()
            print(Fore.RED + err + Fore.RESET)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2023 Aurélien Bompard
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import asyncio
import time

from attr import dataclass

from maubot.matrix import MaubotMatrixClient, MaubotMessageEvent
from mautrix.api import HTTPAPI
from mautrix.types import (
    EventContent,
    EventType,
    MessageEvent,
    MessageType,
    RoomID,
    TextMessageEventContent,
)


@dataclass
class MatrixEvent:
    room_id: RoomID
    event_type: EventType
    content: EventContent
    kwargs: dict


class TestBot:
    """A mocked bot used for testing purposes.

    Send messages to the mock Matrix server with the ``send()`` method.
    Look into the ``responded`` list to get what server has replied.
    """

    def __init__(self, mxid="@botname:example.com", mxurl="http://matrix.example.com"):
        api = HTTPAPI(base_url=mxurl)
        self.client = MaubotMatrixClient(api=api)
        self.responded = []
        self.client.mxid = mxid
        self.client.send_message_event = self._mock_send_message_event

    async def _mock_send_message_event(self, room_id, event_type, content, txn_id=None, **kwargs):
        self.responded.append(
            MatrixEvent(room_id=room_id, event_type=event_type, content=content, kwargs=kwargs)
        )

    async def dispatch(self, event_type: EventType, event):
        tasks = self.client.dispatch_manual_event(event_type, event, force_synchronous=True)
        return await asyncio.gather(*tasks)

    async def send(
        self,
        content,
        html=None,
        room_id="testroom",
        msg_type=MessageType.TEXT,
        sender="@dummy:example.com",
        timestamp=None,
    ):
        event = make_message(
            content,
            html=html,
            room_id=room_id,
            msg_type=msg_type,
            sender=sender,
            timestamp=timestamp,
        )
        await self.dispatch(EventType.ROOM_MESSAGE, MaubotMessageEvent(event, self.client))


def make_message(
    content,
    html=None,
    room_id="testroom",
    msg_type=MessageType.TEXT,
    sender="@dummy:example.com",
    timestamp=None,
):
    """Make a Matrix message event."""
    return MessageEvent(
        type=EventType.ROOM_MESSAGE,
        room_id=room_id,
        event_id="test",
        sender=sender,
        timestamp=timestamp or int(time.time() * 1000),
        content=TextMessageEventContent(msgtype=msg_type, body=content, formatted_body=html),
    )



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import os
import os.path

from ..loader import BasePluginLoader, PluginMeta


class FileSystemLoader(BasePluginLoader):
    def __init__(self, path: str, meta: PluginMeta) -> None:
        self.path = path
        self.meta = meta

    @property
    def source(self) -> str:
        return self.path

    def sync_read_file(self, path: str) -> bytes:
        with open(os.path.join(self.path, path), "rb") as file:
            return file.read()

    async def read_file(self, path: str) -> bytes:
        return self.sync_read_file(path)

    def sync_list_files(self, directory: str) -> list[str]:
        return os.listdir(os.path.join(self.path, directory))

    async def list_files(self, directory: str) -> list[str]:
        return self.sync_list_files(directory)



# maubot - A plugin-based Matrix bot system.
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar
from abc import ABC, abstractmethod
import asyncio

from ..plugin_base import Plugin
from .meta import PluginMeta

if TYPE_CHECKING:
    from ..instance import PluginInstance

PluginClass = TypeVar("PluginClass", bound=Plugin)


class IDConflictError(Exception):
    pass


class BasePluginLoader(ABC):
    meta: PluginMeta

    @property
    @abstractmethod
    def source(self) -> str:
        pass

    def sync_read_file(self, path: str) -> bytes:
        raise NotImplementedError("This loader doesn't support synchronous operations")

    @abstractmethod
    async def read_file(self, path: str) -> bytes:
        pass

    def sync_list_files(self, directory: str) -> list[str]:
        raise NotImplementedError("This loader doesn't support synchronous operations")

    @abstractmethod
    async def list_files(self, directory: str) -> list[str]:
        pass


class PluginLoader(BasePluginLoader, ABC):
    id_cache: dict[str, PluginLoader] = {}

    meta: PluginMeta
    references: set[PluginInstance]

    def __init__(self):
        self.references = set()

    @classmethod
    def find(cls, plugin_id: str) -> PluginLoader:
        return cls.id_cache[plugin_id]

    def to_dict(self) -> dict:
        return {
            "id": self.meta.id,
            "version": str(self.meta.version),
            "instances": [instance.to_dict() for instance in self.references],
        }

    async def stop_instances(self) -> None:
        await asyncio.gather(
            *[instance.stop() for instance in self.references if instance.started]
        )

    async def start_instances(self) -> None:
        await asyncio.gather(
            *[instance.start() for instance in self.references if instance.enabled]
        )

    @abstractmethod
    async def load(self) -> type[PluginClass]:
        pass

    @abstractmethod
    async def reload(self) -> type[PluginClass]:
        pass

    @abstractmethod
    async def delete(self) -> None:
        pass
